import typing as t


def to_bool(val: None | bool | int | float | str) -> bool:
    if val is None:
        return False
    if isinstance(val, bool):
        return val
    if isinstance(val, int):
        if val == 0:
            return False
        return True
    if isinstance(val, float):
        if val == 0.0:
            return False
        return True
    if isinstance(val, str):
        val = val.strip().lower()
        if val == "" or val in ["false", "no", "n", "0"]:
            return False
        if val in ["true", "yes", "y", "1"]:
            return True

    raise ValueError(f"Unable to coerce value to boolean: {val}")



import base64
import binascii

try:
    from kubernetes import client
    from kubernetes import config

except ImportError:
    # TODO: Improve handling of kubernetes as an optional dep
    # kubernetes = None
    # client = None
    # config = None
    pass

from kvcommon.exceptions import KVCException
from kvcommon.logger import get_logger

LOG = get_logger("kvc-k8s")


class KubernetesException(KVCException):
    pass


def get_k8s_secret(secret_name: str, namespace: str):
    """
    Returns a 'v1Secret' struct from kubernetes library

    Raises: 'KubernetesException' on failure
    """
    config.load_incluster_config()
    v1api = client.CoreV1Api()

    LOG.debug("Retrieving K8s secret '%s' in namespace '%s'", secret_name, namespace)

    v1secret = v1api.read_namespaced_secret(secret_name, namespace)
    if not v1secret or not hasattr(v1secret, "data"):
        raise KubernetesException(
            f"No secret retrieved for name: '{secret_name}' in namespace: '{namespace}'"
        )
    return v1secret


def get_secret_strings(secret) -> dict:
    """
    Get dict of b64-decoded strings from opaque K8s V1Secret data dict
    """

    # TODO: Some kind of typehinting wrapping of kubernetes lib
    if not secret or not hasattr(secret, "data"):
        raise KubernetesException(f"Invalid secret")

    secret_data: dict = secret.data
    data_dict = dict()

    for key, value in secret_data.items():
        try:
            data_dict[key] = base64.b64decode(value).decode("utf-8")
        except binascii.error as ex:  # type: ignore
            raise KubernetesException(
                f"Failed to b64decode value at key: '{key}' in data for secret with name: '{value}'.\n"
                f"base64 exception: {ex}"
            )

    return data_dict


def get_decoded_k8s_secret(secret_name: str, namespace: str) -> dict:
    """
    Gets a 'v1Secret' struct from kubernetes and decodes

    Raises: 'KubernetesException' on failure
    """
    try:
        secret = get_k8s_secret(secret_name=secret_name, namespace=namespace)
    except KubernetesException:
        # TODO
        raise

    return get_secret_strings(secret)



import logging
import typing as t

from .vars import KVC_LOG_DATEFMT
from .vars import KVC_LOG_FORMAT

logging_format_string = KVC_LOG_FORMAT or "%(asctime)s - [%(levelname)s] - %(name)s - %(message)s"
logging_format_time = KVC_LOG_DATEFMT or "[%Y-%m-%d %H:%M:%S]"


def get_logger(
    name: t.Optional[str] = None,
    console_log_level=logging.DEBUG,
    logging_format_string: str = logging_format_string,
    logging_format_time: str = logging_format_time,
    filters: t.Iterable[logging.Filter] | None = None
):
    logger = logging.getLogger(name=name)
    logger.setLevel(console_log_level)

    formatter = logging.Formatter(logging_format_string, logging_format_time)

    ch = logging.StreamHandler()
    ch.setLevel(console_log_level)
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    if filters is not None:
        for filter in filters:
            logger.addFilter(filter)

    return logger



from __future__ import annotations
from abc import ABC
from abc import abstractmethod
import os
import pathlib
import typing as t

import toml

from kvcommon import logger



LOG = logger.get_logger("kvc-ds")


class DatastoreBackend(ABC):
    @property
    @abstractmethod
    def _data_by_ref(self) -> dict:
        """
        Return a reference to a dict containing this backend's data
        """
        raise NotImplementedError()

    @property
    def _data(self) -> dict:
        """
        Return a copy of the the dict containing this backend's data
        """
        return self._data_by_ref.copy()

    def get(self, key, default=None, by_ref: bool = False):
        """
        Get a key:value pair in the backend.
        Access the backend's data by-reference (skipping the dict.copy() call) if by_ref is True.
        """
        if by_ref:
            return self._data_by_ref.get(key, default)
        return self._data.get(key, default)

    @abstractmethod
    def set(self, key, value) -> None:
        """
        Set a key:value pair in the backend
        """
        raise NotImplementedError()

    @abstractmethod
    def overwrite_data(self, data: dict) -> None:
        """
        Overwrite the entire contents of the backend's data with a new dict
        """
        raise NotImplementedError()

    def update_data(self, **overrides: dict) -> None:
        """
        Update the backend's data dict with the contents of the overrides dict
        """
        data = self._data_by_ref
        data.update(overrides)
        self.overwrite_data(data)


class DictBackend(DatastoreBackend):
    _data_dict: dict

    def __init__(self) -> None:
        self._data_dict = dict()

    @property
    def _data_by_ref(self) -> dict:
        return self._data_dict

    def set(self, key, value):
        self._data_dict[key] = value

    def overwrite_data(self, data: dict) -> None:
        self._data_dict = data


class TOMLBackend(DatastoreBackend):
    def __init__(
        self,
        storage_dir_path: str | pathlib.Path,
        user_conf_filename: str | pathlib.Path,
    ) -> None:
        self.USER_DIR = pathlib.Path(storage_dir_path)
        self.USER_CONF_PATH = self.USER_DIR / user_conf_filename
        if not str(self.USER_CONF_PATH).endswith(".toml"):
            self.USER_CONF_PATH = self.USER_CONF_PATH.with_suffix(".toml")

    @property
    def _data_by_ref(self) -> dict:
        return self.read_data()

    def set(self, key, value):
        self.update_data(**{key: value})

    def overwrite_data(self, data: dict) -> None:
        self.write_data(data)

    def write_data(self, data: dict):
        """Writes a config dict as a .toml file to the path from constants.

        Also ensure that permissions are appropriate for a file that we're storing
        tokens and secrets in.
        """
        try:
            # TODO: Better error handling
            self.USER_DIR.mkdir(exist_ok=True, parents=True)
            self.USER_CONF_PATH.touch(exist_ok=True)

            # Ensure permissions are good before we store anything
            os.chmod(self.USER_DIR, 0o700)
            os.chmod(self.USER_CONF_PATH, 0o700)

            with open(self.USER_CONF_PATH, "w") as toml_file:
                toml.dump(data, toml_file)
        except OSError as ex:
            LOG.error(
                "Failed to write config to TOML at path: '%s' - full_exception=%s",
                self.USER_CONF_PATH,
                ex,
            )
            raise ex

    def read_data(self) -> dict:
        """Load the config .toml file from its path in the user dir.

        Returns the user config as a dict. If the file doesn't already exist,
        a fresh config file is created and its contents returned.
        """
        data = dict()
        if not self.USER_CONF_PATH.is_file():
            self.write_data(data)

        else:
            try:
                data: dict = toml.load(self.USER_CONF_PATH)
            except OSError as ex:
                LOG.error(
                    "Failed to load config from TOML at path: '%s' - full_exception=%s",
                    self.USER_CONF_PATH,
                    ex,
                )
                raise ex

        return data


# TODO
class YAMLBackend(DatastoreBackend):
    def __init__(self) -> None:
        raise NotImplementedError()
        super().__init__()



import os
import typing as t

KVC_LOG_FORMAT: str = os.getenv(
    "KVC_LOG_FORMAT", "%(asctime)s - [%(levelname)s] - %(name)s - %(message)s"
)
KVC_LOG_DATEFMT: str = os.getenv("KVC_LOG_DATEFMT", "%Y-%m-%d %H:%M:%S")



import typing as t
import subprocess

from kvcommon.logger import get_logger

logger = get_logger("KVC")


def shell_cmd(cmd: str, raise_exceptions: bool = False) -> str:
    try:
        command_line_process = subprocess.Popen(
            cmd,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )

        stdout, stderr = command_line_process.communicate()

        if stderr:
            logger.error(stderr)

        if isinstance(stdout, str):
            return stdout
        return stdout.decode("utf-8")

    except (OSError, subprocess.CalledProcessError) as ex:
        logger.error("Exception occured: " + str(ex))
        if raise_exceptions:
            raise(ex)
        return str(ex)



from __future__ import annotations
import typing as t

from kvcommon import logger

from .backend import DatastoreBackend


LOG = logger.get_logger("kvc-ds")


class VersionedDatastore(object):
    """
    Versioned Datastore

    An over-engineered wrapper around dict storage with rudimentary versioning.
    """

    _config_version: int
    _version_key = "_data_version"
    _backend: DatastoreBackend

    def __init__(
        self, backend: DatastoreBackend | t.Type[DatastoreBackend], config_version: int
    ) -> None:
        self._config_version = config_version
        if isinstance(backend, DatastoreBackend):
            self._backend = backend
        elif isinstance(backend, type(DatastoreBackend)):
            self._backend = backend()

        if not self.check_version_match():
            self._migrate_version()

    def check_version_match(self) -> bool:
        return self.get_data_version(allow_none=True) == self._config_version

    def _migrate_version(self):
        LOG.debug("New config version detected. Migrating.")
        # ...
        # Subclasses should super() call this method after migrating data
        # ...
        self._set_data_version(self._config_version)

    def _set_data_version(self, value: int):
        if not isinstance(value, int):
            raise TypeError(f"Attempting to set non-int value for config_version: {value}")
        self.set_value(self._version_key, value)

    def get_data_version(self, allow_none=True) -> int:
        value = self.get_value(self._version_key)
        if allow_none and value is None:
            return 0
        if not isinstance(value, int):
            raise TypeError(f"Non-int value retrieved for config_version: {value}")
        return value

    def get_value(self, key, default=None, by_ref=False) -> t.Any:
        return self._backend.get(key, default=default, by_ref=by_ref)

    def set_value(self, key, value):
        return self._backend.set(key, value)

    def get_or_create_nested_dict(self, key) -> dict:
        existing = self.get_value(key, default=None)
        if existing is None:
            existing = {}
            self.set_value(key, existing)
        elif not isinstance(existing, dict):
            raise TypeError(f"Non-dict value retrieved for key: {key}")
        return existing

    def overwrite_data(self, new_data: dict) -> None:
        return self._backend.overwrite_data(new_data)

    def update_data(self, **overrides: dict) -> None:
        return self._backend.update_data(**overrides)



import pprint


def pp(
    *args,
    pprint_indent: int = 4,
    pprint_width: int = 100,
    pprint_depth: int | None = None,
    pprint_compact: bool = True,
    pprint_sort_dicts: bool = True,
    pprint_stream=None,
    **kwargs,
) -> pprint.PrettyPrinter:
    """
    PrettyPrinter is nice, but involves way too much boilerplate in setting one up quickly.

    This function acts as a one-liner that instantiates a pprint.PrettyPrinter with my preferred defaults and invokes it.

    The PrettyPrinter object is returned for later re-use.

    Params changed from defaults:
        pprint_compact: False -> True
        pprint_indent: 1 -> 4
        pprint_width: 80 -> 100
    """
    pp = pprint.PrettyPrinter(
        indent=pprint_indent,
        width=pprint_width,
        depth=pprint_depth,
        sort_dicts=pprint_sort_dicts,
        compact=pprint_compact,
        stream=pprint_stream,
    )
    pp.pprint(*args, **kwargs)
    return pp



import typing as t
from urllib.parse import urlparse
from urllib.parse import ParseResult


def urlparse_ignore_scheme(
    url: str, scheme: str = "http", force_scheme: bool = False
) -> ParseResult:
    """This function performs urllib.parse.urlparse on a URL string while ignoring its scheme.

    This allows for calling urlparse on a URL that is lacking a scheme (e.g. to extract the netloc),
    or with an alternate scheme.
    This function sets a scheme if one is specified or absent, to satisfy urlparse's expectations.

    Args:
        scheme: Scheme to use for urlparse. Defaults to "http"
        force_scheme: Ensures the scheme parsed-against is returned in the ParseResult, else "" is returned for scheme
    """

    if not scheme:
        scheme = "http"

    url_parts = None
    no_input_scheme = False

    if url.startswith("://"):
        url = f"{scheme}://{url[3:]}"
        no_input_scheme = True

    if "://" not in url[0:12]:
        url = f"{scheme}://{url}"
        no_input_scheme = True

    url_parts = urlparse(url)

    # Return new ParseResult with blanked scheme since we faked it only for the purposes of making urlparse work
    out_scheme = url_parts.scheme
    if no_input_scheme:
        out_scheme = ""
        if force_scheme:
            out_scheme = scheme

    return ParseResult(
        out_scheme,
        url_parts.netloc,
        url_parts.path,
        url_parts.params,
        url_parts.query,
        url_parts.fragment,
    )


def get_netloc_without_port_from_url_parts(url_parts: ParseResult) -> str:
    # urlparse sadly doesn't provide an easy way to remove/replace the port in netloc
    netloc = url_parts.netloc
    port_str = ":%s" % url_parts.port
    if netloc.endswith(port_str):
        netloc = netloc[: len(netloc) - len(port_str)]
    return netloc
